#!/usr/bin/env python

"""
Provides useful functions for interacting with the oneID API, including creation of
keys, JWTs, etc.
"""

import os
import json
import base64
import re
import time
import logging

try:
    from urllib.request import urlopen, Request
except ImportError:
    from urllib2 import urlopen, Request

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization \
    import Encoding, PrivateFormat, NoEncryption

from .keychain import Keypair
from . import utils

logger = logging.getLogger(__name__)


AUTHENTICATION_ENDPOINT = 'http://developer-portal.oneid.com/api/{project}/authenticate'


ONEID_TYPES = utils.enum(DEVICE=0, SERVER=1, USER=2)

B64_URLSAFE_RE = '[0-9a-zA-Z-_]+'
JWT_RE = r'^{b64}\.{b64}\.{b64}$'.format(b64=B64_URLSAFE_RE)

REQUIRED_JWT_HEADER_ELEMENTS = {
    'typ': 'JWT',
    'alg': 'ES256',
}
TOKEN_EXPIRATION_TIME_SEC = (1*60*60)  # one hour
TOKEN_NOT_BEFORE_LEEWAY_SEC = (2*60)   # two minutes
TOKEN_EXPIRATION_LEEWAY_SEC = (0)      # not really needed


class ServiceCreator(object):
    """
    Read yaml file and add methods dynamically from file
    Created by Session
    """
    def create_service_class(self, service_name, service_model, session, **kwargs):
        """
        Service Model is either user, server or edge_device
        """
        class_attrs = self._create_methods(service_model, **kwargs)
        cls = type(service_name, (BaseService,), class_attrs)

        return cls(session, kwargs.get('project_credentials'))

    def _create_methods(self, service_model, **kwargs):
        """
        :param service_model:
        :return: Dictionary of class attributes
        """
        base_url = kwargs.get('base_url', '')

        methods = dict()
        for method_name, method_values in service_model.iteritems():
            required_jwt = list()
            all_jwt = list()
            for arg_name, arg_properties in method_values['arguments'].iteritems():
                if arg_properties['location'] == 'jwt':
                    all_jwt.append(arg_name)
                    if arg_properties['required'] is True:
                        required_jwt.append(arg_name)

            absolute_url = '{base}{endpoint}'.format(base=base_url,
                                                     endpoint=method_values['endpoint'])

            methods[method_name] = self._create_api_method(method_name,
                                                           absolute_url,
                                                           method_values['method'],
                                                           all_body_args=all_jwt,
                                                           required_body_args=required_jwt,
                                                           )
        return methods

    def _create_api_method(self, name,
                           endpoint, http_method,
                           all_body_args, required_body_args):
        """
        Add methods to session dynamically from yaml file

        :param method_name: method that will be called
        """
        def _api_call(self, *args, **kwargs):
            if kwargs.get('body') is None:
                # if the body isn't specified, check for
                # required body arguments
                for required in required_body_args:
                    if required not in kwargs:
                        raise TypeError('Missing Required Keyword Argument:'
                                        ' %s' % required)
                kwargs.update(body_args=all_body_args)
            return self._make_api_request(endpoint, http_method, **kwargs)

        _api_call.__name__ = name
        return _api_call


class BaseService(object):
    """
    Dynamically loaded by data files.
    """
    def __init__(self, session, project_credentials=None):
        """
        Create a new Service

        :param session: :class:`oneid.session.Session` instance
        """
        self.session = session

        self.project_credentials = None
        if hasattr(self.session, 'project_credentials'):
            self.project_credentials = self.session.project_credentials

        self.identity = self.session.identity_credentials.id
        self.credentials = self.session.identity_credentials

        if self.project_credentials and self.project_credentials.id:
            self.project_id = self.project_credentials.id

    def _format_url(self, url_template, **kwargs):
        """
        Url from yaml may require formatting

        :Example:

            /project/{project_id}
            >>> /project/abc-123

        :param url_template: url with arguments that need replaced by vars
        :param params: Dictionary lookup to replace url arguments with
        :return: absolute url
        """
        encoded_params = dict()
        url_args = re.findall(r'{(\w+)}', url_template)
        for url_arg in url_args:
            if url_arg in kwargs:
                encoded_params[url_arg] = kwargs[url_arg]
            elif hasattr(self, url_arg):
                # Check if the argument is a class attribute (i.e. project_id)
                encoded_params[url_arg] = getattr(self, url_arg)
            else:
                raise TypeError('Missing URL argument %s' % url_arg)
        return url_template.format(**encoded_params)

    def _make_api_request(self, endpoint, http_method, **kwargs):
        """
        Convenience method to make HTTP requests and handle responses/error codes

        :param endpoint: URL to the resource
        :param http_method: HTTP method, GET, POST, PUT, DELETE
        :param kwargs: Params to pass to the body or url
        """
        # Split the params based on their type (url or jwt)
        url = self._format_url(endpoint, **kwargs)

        if kwargs.get('body_args'):
            additional_claims = dict()
            for body in kwargs.get('body_args'):
                additional_claims[body] = kwargs[body]

            payload = self.session.create_jwt_payload(**additional_claims)
            jwt = '{payload}.{signature}'.format(payload=payload,
                                                 signature=self.credentials.keypair.sign(payload))
            return self.session.service_request(http_method, url, body=jwt)
        elif kwargs.get('body'):
            # Replace the entire body with kwargs['body']
            return self.session.service_request(http_method, url,
                                                body=kwargs.get('body'))
        else:
            return self.session.service_request(http_method, url)


def create_secret_key(output=None):
    """
    Create a secret key and save it to a secure location

    :param output: Path to save the secret key
    :return: oneid.keychain.Keypair
    """
    secret_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
    secret_key_bytes = secret_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption())

    # Save the secret key bytes to a secure file
    if output and os.path.exists(os.path.dirname(output)):
        with open(output, 'w') as f:
            f.write(secret_key_bytes)

    return Keypair.from_secret_pem(key_bytes=secret_key_bytes)


def create_aes_key():
    """
    Create an AES256 key for symmetric encryption

    :return: Encryption key bytes
    """
    return os.urandom(32)


def encrypt_attr_value(attr_value, aes_key):
    """
    Convenience method to encrypt attribute properties

    :param attr_value: plain text that you want encrypted
    :param aes_key: symmetric key to encrypt attribute value with
    :return: Dictionary with base64 encoded cipher text and base 64 encoded iv
    """
    iv = os.urandom(16)
    cipher_alg = Cipher(algorithms.AES(aes_key), modes.GCM(iv), backend=default_backend())
    encryptor = cipher_alg.encryptor()
    encr_value = encryptor.update(attr_value) + encryptor.finalize()
    encr_value_b64 = base64.b64encode(encr_value + encryptor.tag)
    iv_b64 = base64.b64encode(iv)
    return {'cipher': 'aes', 'mode': 'gcm', 'ts': 128, 'iv': iv_b64, 'ct': encr_value_b64}


def decrypt_attr_value(attr_ct, aes_key):
    """
    Convenience method to decrypt attribute properties

    :param attr_ct: Dictionary with base64 encoded cipher text and base 64 encoded iv
    :param aes_key: symmetric key to decrypt attribute value with
    :return: Dictionary with base64 encoded cipher text and base 64 encoded iv
    """
    if not isinstance(attr_ct, dict) or attr_ct.get('cipher', 'aes') != 'aes' or attr_ct.get('mode', 'gcm') != 'gcm':
        raise ValueError('invalid encrypted attribute')
    iv = base64.b64decode(attr_ct['iv'])
    tag_ct = base64.b64decode(attr_ct['ct'])
    ts = attr_ct.get('ts', 64) // 8
    tag = tag_ct[-ts:]
    ct = tag_ct[:-ts]
    cipher_alg = Cipher(algorithms.AES(aes_key), modes.GCM(iv, tag, min_tag_length=8), backend=default_backend())
    decryptor = cipher_alg.decryptor()
    return decryptor.update(ct) + decryptor.finalize()


def make_jwt(claims, authorized_keypair):
    """
    Convert claims into JWT

    :type claims: Dictionary that will be converted to json
    :param claims: payload data
    :param authorized_key: :py:class:`~oneid.keychain.Keypair` to sign the request
    :return: JWT
    """
    alg = {'alg': 'ES256',
           'typ': 'JWT'}
    alg_serialized = json.dumps(alg)
    alg_b64 = utils.base64url_encode(alg_serialized)

    claims_serialized = json.dumps(claims)
    claims_b64 = utils.base64url_encode(claims_serialized)

    payload = '{alg}.{claims}'.format(alg=alg_b64, claims=claims_b64)

    signature = authorized_keypair.sign(payload)

    return '{payload}.{sig}'.format(payload=payload, sig=signature)


def verify_jwt(jwt, verification_keypair=None):  # TODO: require verification_token
    """
    Convert a JWT back to it's claims, if validated by the :py:class:`~oneid.keychain.Token`

    :param jwt: JWT to verify and convert
    :type jwt: str
    :param verification_token: :py:class:`~oneid.keychain.Token` to verify the JWT
    :type param: :py:class:`~oneid.keychain.Token`
    """
    if not re.match(JWT_RE, jwt):
        logger.debug('Given JWT doesnt match pattern: %s', jwt)
        return False

    try:
        header, payload, signature = [utils.base64url_decode(p) for p in jwt.split('.')]
    except:
        logger.debug('invalid message, error splitting/decoding: %s', jwt, exc_info=True)
        return False

    if not _verify_jwt_header(header.decode('utf-8')):
        return False

    message = _verify_jwt_claims(payload)

    if message is None:
        logger.debug('no message: %s', message)
        return False

    if verification_keypair and not verification_keypair.verify(*(str(jwt).rsplit('.', 1))):
        logger.debug('invalid signature, header=%s, message=%s', header, message)
        return False

    return message


def _verify_jwt_header(header):
    try:
        header = json.loads(header)
    except ValueError:
        logger.debug('invalid header, not valid json: %s', header)
        return False
    except Exception:  # pragma: no cover
        logger.debug('unknown error verifying header: %s', header, exc_info=True)
        return False

    for key, value in REQUIRED_JWT_HEADER_ELEMENTS.items():
        if header.pop(key, None) != value:
            logger.debug('invalid header, missing or incorrect %s: %s', key, header)
            return False

    if len(header) > 0:
        logger.debug('invalid header, extra elements: %s', header)
        return False

    return True


def _verify_jwt_claims(payload):
    try:
        message = json.loads(payload)
        now = int(time.time())

        if 'exp' in message and (int(message['exp']) + TOKEN_EXPIRATION_LEEWAY_SEC) < now:
            logger.warning('Expired token, exp=%s, now=%s', message['exp'], now)
            return None

        if 'nbf' in message and (int(message['nbf']) - TOKEN_NOT_BEFORE_LEEWAY_SEC) > now:
            logger.warning('Early token, nbf=%s, now=%s', message['nbf'], now)
            return None

        if 'jti' in message and not utils.verify_and_burn_nonce(message['jti']):
            logger.warning('Invalid nonce: %s', message['jti'])
            return None

        return message

    except:
        logger.debug('unknown error verifying payload: %s', payload, exc_info=True)
        return None





"""
Token is short for key pair, used to sign and verify signatures

Keys should be kept in a secure storage enclave.
"""
import os

import binascii
import base64

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey, \
    EllipticCurvePrivateKey, EllipticCurvePublicNumbers, SECP256R1
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key, \
    load_der_private_key, load_der_public_key

from cryptography.hazmat.primitives.asymmetric.utils \
    import decode_dss_signature, encode_dss_signature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization \
    import Encoding, PublicFormat, PrivateFormat, NoEncryption

from . import utils


class Credentials(object):
    """
    Container for User/Server/Device Encryption Key, Signing Key, Identity


    :ivar identity: UUID of the identity.
    :ivar keypair: :class:`~oneid.keychain.Keypair` instance.
    """
    def __init__(self, uuid, keypair):
        """

        :param identity: uuid of the entity
        :param keypair: :py:class:`~oneid.keychain.Keypair` instance
        """
        self.id = uuid

        if not isinstance(keypair, Keypair):
            raise ValueError('keypair must be a oneid.keychain.Keypair instance')

        self.keypair = keypair


class ProjectCredentials(Credentials):
    def __init__(self, uuid, keypair, encryption_key):
        """
        Adds an ecryption key

        :param uuid: oneID project UUID
        :param keypair: :py:class:`~oneid.keychain.Keypair`
        :param encryption_key: AES key used to encrypt messages
        """
        super(ProjectCredentials, self).__init__(uuid, keypair)
        self._encryption_key = encryption_key

    def encrypt(self, plain_text):
        """
        Encrypt plain text with the project encryption key.

        :param plain_text: String to encrypt with project encryption key.
        :returns: Dictionary with cipher text and encryption params.
        """
        iv = os.urandom(16)
        cipher_alg = Cipher(algorithms.AES(self._encryption_key), modes.GCM(iv), backend=default_backend())
        encryptor = cipher_alg.encryptor()
        encr_value = encryptor.update(plain_text) + encryptor.finalize()
        encr_value_b64 = base64.b64encode(encr_value + encryptor.tag)
        iv_b64 = base64.b64encode(iv)
        return {'cipher': 'aes', 'mode': 'gcm', 'ts': 128, 'iv': iv_b64, 'ct': encr_value_b64}

    def decrypt(self, cipher_text, iv=None, cipher='aes', mode='gcm', tag_size=128):
        """
        Decrypt cipher text that was encrypted with the project encryption key

        :param cipher_text: Encrypted text
        :param iv: Base64 encoded initialization vector
        :param mode: encryption mode i.e. gcm
        :param tag_size: tag size, default 128
        :returns: plain text
        """
        if cipher.lower() == 'aes' and mode.lower() == 'gcm' and iv is None:
            raise ValueError('IV must be specified with using AES and GCM')

        iv = base64.b64decode(iv)
        tag_ct = base64.b64decode(cipher_text)
        ts = tag_size // 8
        tag = tag_ct[-ts:]
        ct = tag_ct[:-ts]
        cipher_alg = Cipher(algorithms.AES(self._encryption_key),
                            modes.GCM(iv, tag, min_tag_length=8),
                            backend=default_backend())
        decryptor = cipher_alg.decryptor()
        return decryptor.update(ct) + decryptor.finalize()


class Keypair(object):
    def __init__(self, *args, **kwargs):
        """
        :param kwargs: Pass secret key bytes
        """
        self.identity = kwargs.get('identity')

        self._private_key = None
        self._public_key = None

        if kwargs.get('secret_bytes') and \
                isinstance(kwargs['secret_bytes'], EllipticCurvePrivateKey):
            self._load_secret_bytes(kwargs['secret_bytes'])

    def _load_secret_bytes(self, secret_bytes):
        self._private_key = secret_bytes

    @property
    def secret_as_der(self):
        """
        Write out the private key as a DER format

        :return: DER encoded private key
        """
        secret_der = self._private_key.private_bytes(Encoding.DER, PrivateFormat.PKCS8, NoEncryption())

        return secret_der

    @property
    def secret_as_pem(self):
        """
        Write out the private key as a PEM format

        :return: Pem Encoded private key
        """
        return self._private_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption())

    @classmethod
    def from_secret_pem(cls, key_bytes=None, path=None):
        """
        Create a :class:`~oneid.keychain.Keypair` from a pem formatted data

        :return: :class:`~oneid.keychain.Keypair` instance
        """
        if key_bytes:
            secret_bytes = load_pem_private_key(key_bytes, None, default_backend())
            return cls(secret_bytes=secret_bytes)

        if os.path.exists(path):
            with open(path, 'r') as pem_file:
                secret_bytes = load_pem_private_key(pem_file.read(), None, default_backend())
                return cls(secret_bytes=secret_bytes)

    @classmethod
    def from_secret_der(cls, der_key):
        """
        Read a der_key, convert it a private key
        :param path: der formatted key
        :return:
        """
        secret_bytes = load_der_private_key(der_key, None, default_backend())
        return cls(secret_bytes=secret_bytes)

    def _load_public_key_by_coord(self, x, y):
        """
        load validate key by the curve points
        :param x: long x coordinate of ecc curve
        :param y: long y coordinate of ecc curve
        :return:
        """
        self._public_key = EllipticCurvePublicNumbers(x, y, SECP256R1()).public_key(default_backend())

    @classmethod
    def from_public_der(cls, public_key):
        """
        Given a DER-format public key, convert it into a token to
        validate signatures
        :param public_key: der formatted key
        :return: :class:`~oneid.keychain.Keypair` instace
        """
        pub = load_der_public_key(public_key, default_backend())

        new_token = cls()
        new_token._public_key = pub

        return new_token

    def verify(self, payload, signature):
        """
        Verify that the token signed the data
        :type payload: String
        :param payload: message that was signed and needs verified
        :type signature: Base64 URL Safe
        :param signature: Signature that can verify the senders
         identity and payload
        :return:
        """
        raw_sig = utils.base64url_decode(signature)
        sig_r_bin = raw_sig[:len(raw_sig)/2]
        sig_s_bin = raw_sig[len(raw_sig)/2:]

        sig_r = unpack_bytes(sig_r_bin)
        sig_s = unpack_bytes(sig_s_bin)

        sig = encode_dss_signature(sig_r, sig_s)
        signer = self.public_key.verifier(sig,
                                          ec.ECDSA(hashes.SHA256()))
        signer.update(payload)
        return signer.verify()

    def sign(self, payload):
        """
        Sign a payload
        :param payload: String (usually jwt payload)
        :return: URL safe base64 signature
        """
        signer = self._private_key.signer(ec.ECDSA(hashes.SHA256()))

        signer.update(payload)
        signature = signer.finalize()

        r, s = decode_dss_signature(signature)

        b64_signature = utils.base64url_encode('{r}{s}'.format(r=int2bytes(r),
                                                               s=int2bytes(s)))
        return b64_signature

    @property
    def public_key(self):
        """
        If the private key is defined, generate the public key
        :return:
        """
        if self._public_key:
            return self._public_key
        elif self._private_key:
            return self._private_key.public_key()

    @property
    def public_key_der(self):
        """
        DER formatted public key

        :return: Public Key in DER format
        """
        return self.public_key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)

    @property
    def public_key_pem(self):
        """
        PEM formatted public key

        :return: Public Key in PEM format
        """
        return self.public_key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)

    def save(self, *args, **kwargs):
        """
        Save a key.
        Should be overridden and saved to secure storage

        :param args:
        :param kwargs:
        :return: Bool Success
        """
        raise NotImplementedError


def int2bytes(i):
    hex_string = '%x' % i
    n = len(hex_string)
    return binascii.unhexlify(hex_string.zfill(n + (n & 1)))


def unpack_bytes(stringbytes):
    return int(binascii.hexlify(stringbytes), 16)



import os
import yaml
import json
import base64
from requests import request

from . import service, utils, exceptions

REQUIRED_JWT_HEADER_ELEMENTS = {
    'typ': 'JWT',
    'alg': 'ES256',
}


class SessionBase(object):
    """
    Abstract Session Class

    :ivar identity_credentials: oneID identity :class:`~oneid.keychain.Credentials`
    :ivar application_credentials: unique app credentials :class:`~oneid.keychain.Credentials`
    :ivar project_credentials: unique project credentials :class:`~oneid.keychain.Credentials`
    :ivar oneid_credentials: oneID project credentials :class:`~oneid.keychain.Credentials`
    """
    def __init__(self, identity_credentials=None, application_credentials=None,
                 project_credentials=None, oneid_credentials=None, config=None):
        """

        :param identity_credentials: :py:class:`~oneid.keychain.Credentials`
        :param application_credentials: :py:class:`~oneid.keychain.Credentials`
        :param project_credentials: :py:class:`~oneid.keychain.ProjectCredentials`
        :param oneid_credentials: :py:class:`~oneid.keychain.Credentials`
        :param config: Dictionary or configuration keyword arguments
        :return:
        """
        self.identity_credentials = identity_credentials
        self.app_credentials = application_credentials
        self.project_credentials = project_credentials
        self.oneid_credentials = oneid_credentials

    def _load_config(self, config_file):
        """
        Load configuration from file
        :return: dict()
        """
        # Load params from configuration file
        with open(config_file, mode='r') as config:
            params = yaml.safe_load(config)
            return params

    def _create_services(self, methods, **kwargs):
        """
        Populate session variables and create methods from args

        :return: None
        """
        service_creator = service.ServiceCreator()

        for method in methods:
            if method != 'GLOBAL':
                setattr(self, method,
                        service_creator.create_service_class(method,
                                                             methods[method],
                                                             self,
                                                             **kwargs)
                        )

    def create_jwt_payload(self, **kwargs):
        """
        Create a generic JWT Payload

        :param claims: JWT Claims Dict()
        :return: JWT payload (*no signature)
        """
        alg_b64 = base64.b64encode(json.dumps(REQUIRED_JWT_HEADER_ELEMENTS))

        # Required claims
        jti = utils.make_nonce()

        claims = {'jti': jti}
        claims.update(kwargs)

        claims_serialized = json.dumps(claims)
        claims_b64 = base64.b64encode(claims_serialized)

        payload = '{alg_b64}.{claims_b64}'.format(alg_b64=alg_b64,
                                                  claims_b64=claims_b64)

        return payload

    def make_http_request(self, http_method, url,
                           headers=None, body=None):
        """
        Generic HTTP request

        :param headers:
        :param body:
        :return:
        """
        valid_http_methods = ['GET', 'PUT', 'POST', 'DELETE']
        if http_method not in valid_http_methods:
            raise TypeError('HTTP method must be %s' %
                            ', '.join(valid_http_methods))

        req = request(http_method, url, headers=headers, data=body)

        # 403 is Forbidden, raise an error if this occurs
        if req.status_code == 403:
            raise exceptions.InvalidAuthentication()

        return req.content

    def prepare_message(self, *args, **kwargs):
        raise NotImplementedError

    def send_message(self, *args, **kwargs):
        raise NotImplementedError

    def verify_message(self, *args, **kwargs):
        raise NotImplementedError


class DeviceSession(SessionBase):
    def __init__(self, identity_credentials=None, application_credentials=None,
                 project_credentials=None, oneid_credentials=None, config=None):
        super(DeviceSession, self).__init__(identity_credentials,
                                            application_credentials,
                                            project_credentials,
                                            oneid_credentials, config)

    def verify_message(self, message):
        """
        Verify a message received from the server

        :param message: JSON formatted message with two signatures
        :return: verified message
        :raises: oneid.exceptions.InvalidAuthentication
        """
        data = json.loads(message)
        if not data.get('payload'):
            raise KeyError('missing payload')
        if not data.get('oneid_signature'):
            raise KeyError('missing oneID Digital Signature')
        if not data.get('project_signature'):
            raise KeyError('missing project signature')

        # Verify the signatures
        payload = data['payload'].encode('utf-8')
        project_sig = data['project_signature']
        oneid_sig = data['oneid_signature']

        self.project_credentials.keypair.verify(payload, project_sig)
        self.oneid_credentials.keypair.verify(payload, oneid_sig)

    def prepare_message(self, **kwargs):
        """
        Prepare a message before sending

        :return: JSON with JWT payload and two signatures
        """
        kwargs['iss'] = self.identity_credentials.id
        payload = self.create_jwt_payload(**kwargs)
        identity_sig = self.identity_credentials.keypair.sign(payload)
        app_sig = self.app_credentials.keypair.sign(payload)

        return json.dumps({'payload': payload,
                           'id_signature': identity_sig,
                           'app_signature': app_sig})

    def send_message(self, *args, **kwargs):
        raise NotImplementedError


class ServerSession(SessionBase):
    """
    Enable Server to request two-factor Authentication from oneID
    """
    def __init__(self, identity_credentials=None, application_credentials=None,
                 project_credentials=None, oneid_credentials=None, config=None):
        super(ServerSession, self).__init__(identity_credentials,
                                            application_credentials,
                                            project_credentials,
                                            oneid_credentials, config)

        if isinstance(config, dict):
            params = config
        else:
            # Load default
            default_config = os.path.join(os.path.dirname(__file__),
                                          'data', 'oneid_server.yaml')
            params = self._load_config(config if config else default_config)

        self._create_services(params)

    def _create_services(self, params, **kwargs):
        """
        Populate session variables and create methods from
        :return: None
        """
        global_kwargs = params.get('GLOBAL', {})
        if self.project_credentials:
            global_kwargs['project_credentials'] = self.project_credentials

        super(ServerSession, self)._create_services(params, **global_kwargs)

    def service_request(self, http_method, endpoint, body=None):
        """
        Make an API Request

        :param method:
        :param endpoint:
        :param body:
        :return:
        """
        payload = self.create_jwt_payload()

        signature = self.identity_credentials.keypair.sign(payload)

        auth_jwt_header = '{payload}.{signature}'.format(payload=payload,
                                                         signature=signature)

        headers = {
            'Content-Type': 'application/jwt',
            'Authorization': 'Bearer %s' % auth_jwt_header
        }

        response = self.make_http_request(http_method, endpoint, headers=headers,
                                          body=body)

        return response

    def prepare_message(self, *args, **kwargs):
        """
        Build message that has two-factor signatures

        :param kwargs: Claims to add to the JWT
        :return: Content to be sent to devices
        """
        if self.project_credentials is None:
            raise AttributeError

        oneid_response = kwargs.pop('oneid_response')
        # split the JWT Token
        alg, claims, oneid_sig = oneid_response.split('.')
        payload = '{alg}.{claims}'.format(alg=alg, claims=claims)

        project_sig = self.project_credentials.keypair.sign(payload)

        return json.dumps({'payload': payload,
                           'project_signature': project_sig,
                           'oneid_signature': oneid_sig})

    def send_message(self, *args, **kwargs):
        raise NotImplementedError

    def verify_message(self, *args, **kwargs):
        raise NotImplementedError


class AdminSession(SessionBase):
    """
    Admin Users will only interface with oneID service,
    They only need an identity_credentials and oneid_credentials
    to verify responses
    """
    def __init__(self, identity_credentials, application_credentials=None,
                 project_credentials=None, oneid_credentials=None, config=None):
        super(AdminSession, self).__init__(identity_credentials,
                                            application_credentials,
                                            project_credentials,
                                            oneid_credentials, config)

        if isinstance(config, dict):
            params = config
        else:
            default_config = os.path.join(os.path.dirname(__file__),
                                          'data', 'oneid_admin.yaml')
            params = self._load_config(config if config else default_config)

        self._create_services(params)

    def _create_services(self, params, **kwargs):
        """
        Populate session variables and create methods from
        :return: None
        """
        global_kwargs = params.get('GLOBAL', {})
        if self.project_credentials:
            global_kwargs['project_credentials'] = self.project_credentials

        super(AdminSession, self)._create_services(params, **global_kwargs)

    def service_request(self, http_method, endpoint, body=None):
        """
        Make an API Request

        :param method:
        :param endpoint:
        :param body:
        :return:
        """
        payload = self.create_jwt_payload()

        signature = self.identity_credentials.keypair.sign(payload)

        auth_jwt_header = '{payload}.{signature}'.format(payload=payload,
                                                         signature=signature)

        headers = {
            'Content-Type': 'application/jwt',
            'Authorization': 'Bearer %s' % auth_jwt_header
        }

        response = self.make_http_request(http_method, endpoint, headers=headers,
                                          body=body)

        return response

    def prepare_message(self, *args, **kwargs):
        raise NotImplementedError

    def send_message(self, *args, **kwargs):
        raise NotImplementedError

    def verify_message(self, *args, **kwargs):
        raise NotImplementedError





"""
Helpful utility functions
"""

import random
import time
import base64
import re
from datetime import datetime, timedelta
from dateutil import parser, tz
import logging

logger = logging.getLogger(__name__)


def enum(**items):
    """
    Create an enumeration

    :Example:

        Animals = enum(DOG=0, CAT=1)

    :param items: Dict() of keys and integer values
    :return: Enum instance
    """
    return type('Enum', (), items)


def base64url_encode(msg):
    """
    Default b64_encode adds padding, jwt spec removes padding
    :param input:
    :return: base64 en
    """
    encoded_input = base64.urlsafe_b64encode(msg)
    stripped_input = encoded_input.replace('=', '')
    return stripped_input


def base64url_decode(msg):
    """
    JWT spec doesn't allow padding characters. base64url_encode removes them,
    base64url_decode, adds them back in before trying to base64 decode the message

    :param msg: URL safe base64 message
    :return: plain text message
    """
    pad = len(msg) % 4
    if pad > 0:
        msg += '=' * (4 - pad)

    return base64.urlsafe_b64decode(str(msg))


def make_nonce():
    """
    Create a nonce with timestamp included

    :return: nonce
    """
    time_format = '%Y-%m-%dT%H:%M:%SZ'
    time_component = time.strftime(time_format, time.gmtime())
    valid_chars = ''

    # iterate over all the aschii characters for a list of all alpha-numeric characters
    for char_index in range(0, 128):
        if chr(char_index).isalpha() or chr(char_index).isalnum():
            valid_chars += chr(char_index)

    random_str = ''
    random_chr = random.SystemRandom()
    for i in xrange(0, 6):
        random_str += random_chr.choice(valid_chars)

    return '001{time_str}{random_str}'.format(time_str=time_component,
                                              random_str=random_str)


def verify_and_burn_nonce(nonce):
    """
    Ensure that the nonce is correct, less than one hour old,
    and not more than two minutes in the future

    Callers should also store used nonces and reject messages
    with previously-used ones.

    :param nonce: Nonce as created with :func:`~oneid.utils.make_nonce`
    :return: True only if nonce meets validation criteria
    :rtype: bool
    """
    ret = re.match(r'^001[2-9][0-9]{3}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])'
                   r'T([01][0-9]|2[0-3])(:[0-5][0-9]){2}Z[A-Za-z0-9]{6}$', nonce)
    if ret:
        date = parser.parse(nonce[3:-6])
        now = datetime.utcnow().replace(tzinfo=tz.tzutc())
        ret = date < (now + timedelta(minutes=2)) and date > (now + timedelta(hours=-1))

    return ret  # TODO: keep a record (at least for the last hour) of burned nonces
